/* Copyright (c) 2016 SpaceToad and the BuildCraft team
 * 
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at https://mozilla.org/MPL/2.0/. */
package buildcraft.lib.delta;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTTagList;
import net.minecraft.network.PacketBuffer;

import net.minecraftforge.common.util.Constants;

import buildcraft.lib.delta.DeltaManager.EnumDeltaMessage;
import buildcraft.lib.delta.DeltaManager.EnumNetworkVisibility;

public class DeltaInt {
    public final String name;
    public final EnumNetworkVisibility visibility;
    /** The static value at the START of each delta If you add a delta entry then this value will change before delta
     * has completed. */
    private int staticStartValue = 0;
    /** The static value at the END of each delta. If you add a delta entry then this value will change after the delta
     * has completed. */
    private int staticEndValue = 0;
    /** The dynamic value, calculated each tick from all the changing deltas. */
    private double dynamicValueLast = 0, dynamicValueThis;
    public final List<DeltaIntEntry> changingEntries = new ArrayList<>();
    private final DeltaManager manager;
    private long tick = 0;

    public DeltaInt(String name, EnumNetworkVisibility visibility, DeltaManager manager) {
        this.name = name;
        this.visibility = visibility;
        this.manager = manager;
    }

    public void tick() {
        Iterator<DeltaIntEntry> iter = changingEntries.iterator();
        while (iter.hasNext()) {
            DeltaIntEntry delta = iter.next();
            if (tick >= delta.endTick) {
                iter.remove();
                staticEndValue += delta.delta;
            } else if (tick >= delta.startTick && !delta.hasStarted) {
                staticStartValue += delta.delta;
                delta.hasStarted = true;
            }
        }
        double dynamic = staticEndValue;
        for (DeltaIntEntry entry : changingEntries) {
            if (tick < entry.startTick) continue;
            long duration = entry.endTick - entry.startTick;
            long elapsed = tick - entry.startTick;
            double interp = elapsed / (double) duration;
            dynamic += entry.delta * interp;
        }

        dynamicValueLast = dynamicValueThis;
        dynamicValueThis = dynamic;
        tick++;
    }

    public double getDynamic(float partialTicks) {
        if (partialTicks <= 0) {
            return dynamicValueLast;
        } else if (partialTicks >= 1) {
            return dynamicValueThis;
        } else {
            double a = dynamicValueLast * (1 - partialTicks);
            double b = dynamicValueThis * partialTicks;
            return a + b;
        }
    }

    public int getStatic(boolean start) {
        return start ? staticStartValue : staticEndValue;
    }

    void receiveData(EnumDeltaMessage type, PacketBuffer buffer) {
        if (type == EnumDeltaMessage.ADD_SINGLE) {
            long start = buffer.readLong();
            long end = buffer.readLong();
            int delta = buffer.readInt();
            DeltaIntEntry entry = new DeltaIntEntry(start + tick, end + tick, delta);
            changingEntries.add(entry);
        } else if (type == EnumDeltaMessage.SET_VALUE) {
            changingEntries.clear();
            int value = buffer.readInt();
            staticStartValue = value;
            staticEndValue = value;
            dynamicValueLast = value;
            dynamicValueThis = value;
        } else if (type == EnumDeltaMessage.CURRENT_STATE) {
            staticStartValue = buffer.readInt();
            staticEndValue = buffer.readInt();
            changingEntries.clear();
            int count = buffer.readUnsignedShort();
            for (int i = 0; i < count; i++) {
                long start = buffer.readLong() + tick;
                long end = buffer.readLong() + tick;
                int delta = buffer.readInt();
                DeltaIntEntry entry = new DeltaIntEntry(start, end, delta);
                entry.hasStarted = buffer.readBoolean();
                changingEntries.add(entry);
            }
        }
    }

    void writeState(PacketBuffer buffer) {
        buffer.writeInt(staticStartValue);
        buffer.writeInt(staticEndValue);
        buffer.writeShort(changingEntries.size());
        for (DeltaIntEntry delta : changingEntries) {
            buffer.writeLong(delta.startTick - tick);
            buffer.writeLong(delta.endTick - tick);
            buffer.writeInt(delta.delta);
            buffer.writeBoolean(delta.hasStarted);
        }
    }

    /** Adds a delta value
     * 
     * @param start
     * @param end
     * @param delta */
    public void addDelta(long start, long end, int delta) {
        DeltaIntEntry entry = new DeltaIntEntry(start + tick, end + tick, delta);
        changingEntries.add(entry);
        manager.sendDeltaMessage(EnumDeltaMessage.ADD_SINGLE, this, (buffer) -> {
            buffer.writeLong(entry.startTick - tick);
            buffer.writeLong(entry.endTick - tick);
            buffer.writeInt(entry.delta);
        });
    }

    /** Forgets all existing deltas and sets the values to the new value.
     * 
     * @param value */
    public void setValue(int value) {
        changingEntries.clear();
        staticStartValue = value;
        staticEndValue = value;
        dynamicValueLast = value;
        dynamicValueThis = value;
        manager.sendDeltaMessage(EnumDeltaMessage.SET_VALUE, this, (buffer) -> buffer.writeInt(value));
    }

    public void readFromNBT(NBTTagCompound nbt) {
        tick = nbt.getLong("tick");
        staticStartValue = nbt.getInteger("static-start");
        staticEndValue = nbt.getInteger("static-end");
        // dynamic is calculated every tick so there is no need to read + write it
        changingEntries.clear();
        NBTTagList list = nbt.getTagList("changing", Constants.NBT.TAG_COMPOUND);
        for (int i = 0; i < list.tagCount(); i++) {
            NBTTagCompound entryNbt = list.getCompoundTagAt(i);
            long start = entryNbt.getLong("start");
            long end = entryNbt.getLong("end");
            int delta = entryNbt.getInteger("delta");
            DeltaIntEntry entry = new DeltaIntEntry(start, end, delta);
            entry.hasStarted = entryNbt.getBoolean("started");
            changingEntries.add(entry);
        }
    }

    public NBTTagCompound writeToNBT() {
        NBTTagCompound nbt = new NBTTagCompound();
        nbt.setLong("tick", tick);
        nbt.setInteger("static-start", staticStartValue);
        nbt.setInteger("static-end", staticEndValue);
        // dynamic is calculated every tick so there is no need to read + write it
        NBTTagList list = new NBTTagList();
        for (DeltaIntEntry entry : changingEntries) {
            NBTTagCompound entryNbt = new NBTTagCompound();
            entryNbt.setLong("start", entry.startTick);
            entryNbt.setLong("end", entry.endTick);
            entryNbt.setInteger("delta", entry.delta);
            entryNbt.setBoolean("started", entry.hasStarted);
            list.appendTag(entryNbt);
        }
        nbt.setTag("changing", list);
        return nbt;
    }

    private static class DeltaIntEntry {
        private boolean hasStarted = false;
        private final long startTick, endTick;
        private final int delta;

        public DeltaIntEntry(long startTick, long endTick, int delta) {
            this.startTick = startTick;
            this.endTick = endTick;
            this.delta = delta;
        }
    }
}
